package org.opengis.cite.geomatics;

import java.net.URI;
import java.util.Set;
import java.util.logging.Logger;

import org.geotoolkit.geometry.Envelopes;
import org.geotoolkit.geometry.ImmutableEnvelope;
import org.geotoolkit.referencing.CRS;
import org.geotoolkit.referencing.GeodeticCalculator;
import org.geotoolkit.referencing.crs.DefaultGeographicCRS;
import org.opengis.geometry.DirectPosition;
import org.opengis.geometry.Envelope;
import org.opengis.geometry.coordinate.Position;
import org.opengis.referencing.ReferenceIdentifier;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.TransformException;
import org.opengis.util.FactoryException;

/**
 * Provides utility methods for using coordinate reference systems and
 * performing coordinate transformations.
 */
public class GeodesyUtils {

    private static final Logger LOGR = Logger.getLogger(GeodesyUtils.class
            .getPackage().getName());

    /**
     * OGC identifier for WGS 84 (geographic 2D)
     */
    public static final String EPSG_4326 = "urn:ogc:def:crs:EPSG::4326";
    /**
     * OGC identifier for WGS 84 (longitude-latitude)
     */
    public static final String OGC_CRS84 = "urn:ogc:def:crs:OGC:1.3:CRS84";

    /**
     * Returns an immutable envelope representing the valid geographic extent of
     * the CRS identified by the given URI reference.
     * 
     * @param crsRef
     *            An absolute URI that identifies a CRS definition.
     * @return An ImmutableEnvelope object.
     * @throws FactoryException
     *             if the CRS reference cannot be resolved to a known
     *             definition.
     */
    public static ImmutableEnvelope getDomainOfValidity(String crsRef)
            throws FactoryException {
        CoordinateReferenceSystem crs = null;
        if (crsRef.equals(OGC_CRS84)) {
            crs = DefaultGeographicCRS.WGS84;
        } else {
            crs = CRS.decode(extractCRSCode(crsRef));
        }
        Envelope areaOfUse = Envelopes.getDomainOfValidity(crs);
        return new ImmutableEnvelope(areaOfUse);
    }

    /**
     * Extracts an abbreviated CRS code value from the given CRS reference.
     * IllegalArgumentException - If the given string is not a valid URI value.
     * 
     * @param crsRef
     *            A URI value that identifies a CRS definition; the 'urn' or
     *            'http' schemes are supported.
     * @return A String of the form "authority:code" (e.g. EPSG:4326).
     * 
     * @see "OGC 09-048r3: Name type specification - definitions - part 1 -
     *      basic name"
     */
    public static String extractCRSCode(String crsRef) {
        URI uri = URI.create(crsRef);
        String[] authorityVerCode = null;
        if (crsRef.startsWith("urn:ogc:def:crs:")) {
            authorityVerCode = uri.toString().split("crs:")[1].split(":");
        } else if (crsRef.startsWith("http://www.opengis.net/def/crs/")) {
            authorityVerCode = uri.toString().split("crs/")[1].split("/");
        } else {
            throw new IllegalArgumentException("Unsupported CRS reference: "
                    + crsRef);
        }
        return authorityVerCode[0] + ":" + authorityVerCode[2];
    }

    /**
     * Returns a standard OGC identifier for the given coordinate reference
     * system (e.g. "urn:ogc:def:crs:EPSG::4326").
     * 
     * @param crs
     *            A {@link CoordinateReferenceSystem} object.
     * @return A String representing a URN value in the 'ogc' namespace; if no
     *         identifier is found an empty String is returned.
     * 
     * @see "OGC 09-048r3: Name type specification - definitions - part 1 -
     *      basic name"
     */
    public static String getCRSIdentifier(CoordinateReferenceSystem crs) {
        StringBuilder crsId = new StringBuilder("urn:ogc:def:crs:");
        Set<ReferenceIdentifier> identifiers = crs.getIdentifiers();
        if (identifiers.isEmpty()) {
            return "";
        }
        ReferenceIdentifier id = identifiers.iterator().next();
        crsId.append(id.getCodeSpace()).append(":");
        // EPSG definitions are not versioned
        if (!id.getCodeSpace().equalsIgnoreCase("EPSG")) {
            crsId.append(id.getVersion());
        }
        crsId.append(':');
        crsId.append(id.getCode());
        return crsId.toString();
    }

    /**
     * Determines the destination position given the azimuth and distance from
     * some starting position.
     * 
     * @param startingPos
     *            The starting position.
     * @param azimuth
     *            The horizontal angle measured clockwise from a meridian.
     * @param distance
     *            The great-circle (orthodromic) distance in the same units as
     *            the ellipsoid axis (e.g. meters for EPSG 4326).
     * @return A DirectPosition representing the destination position (in the
     *         same CRS as the starting position).
     */
    public static DirectPosition calculateDestination(Position startingPos,
            double azimuth, double distance) {
        CoordinateReferenceSystem crs = startingPos.getDirectPosition()
                .getCoordinateReferenceSystem();
        GeodeticCalculator calculator = new GeodeticCalculator(crs);
        DirectPosition destPos = null;
        // calculator only accepts azimuth values in range +- 180
        if (azimuth > 180) {
            azimuth = azimuth - 360;
        } else if (azimuth < -180) {
            azimuth = azimuth + 360;
        }
        try {
            calculator.setStartingPosition(startingPos);
            calculator.setDirection(azimuth, distance);
            destPos = calculator.getDestinationPosition();
        } catch (TransformException te) {
            // Same CRS so this should never arise
            LOGR.fine(te.getMessage());
        }
        return destPos;
    }
}
